#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <gmp.h>
#include <time.h>
#include <math.h>

#include <X11/Xlib.h>
#include <Imlib2.h>


#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif
typedef struct MPF_FILE_CHUNK
{
  int _mp_prec;                 /* Max precision, in number of `mp_limb_t's.
                                   Set by mpf_init and modified by
                                   mpf_set_prec.  The area pointed to by the
                                   _mp_d field contains `prec' + 1 limbs.  */
  int _mp_size;                 /* abs(_mp_size) is the number of limbs the
                                   last field points to.  If _mp_size is
                                   negative this is a negative number.  */
  mp_exp_t _mp_exp;
} MPF_FILE_CHUNK;

typedef struct RESOLUTION
{
  int height, width, n_palette;
  unsigned long precision; 
} RESOLUTION;


typedef struct COMPLEX_MPF
{
  mpf_t re, im;
} COMPLEX_MPF;

// perhaps there should only be one tree structure

struct color_tree_node
{
  struct color_tree_node *left, *right;
  DATA32 data;
  unsigned long color_counter;
};

typedef struct color_tree_node COLOR_TREE_NODE;

#define DIGEST_SIZE 16
#define KEY_SIZE 15

struct digest_tree_node
{
  struct digest_tree_node *left, *right;
  unsigned char digest[DIGEST_SIZE];
};

typedef struct digest_tree_node DIGEST_TREE_NODE;

typedef unsigned short int FRACTAL_VALUE;

/* any other variables that control the program's IO and whatnot  */
/* should go in here */

typedef struct FRACTAL_OPTIONS
{
  long zoom_level;
  long current_zoom_level;
  char *fractal_filename, *image_filename;
  unsigned char random;
} FRACTAL_OPTIONS;

/* all variables that have to do with the generation of the fractal should */
/* go in here */

typedef struct FRACTAL
{
  unsigned char incomplete;
  unsigned long completed_to;

  COMPLEX_MPF top_left, bottom_right;
  int height, width;
  unsigned long precision;
  DATA32 *image_data;
  unsigned long cutoff;

  FRACTAL_OPTIONS options;
} FRACTAL;

typedef struct FILE_FRACTAL
{
  unsigned char incomplete;
  unsigned long completed_to;
  int height, width;
  unsigned long precision;
  unsigned long cutoff;

  long zoom_level;
  long current_zoom_level;
  unsigned char random;
} FILE_FRACTAL;

/* fractal related values */

#define CUTOFF_VALUE 255*255*15
#define ZOOM_LEVEL 10 // 10
#define DOES_NOT_ESCAPE_VALUE 0
#define MAX_VALUE 2

/* mpf related values */

#define LOW_RES_PRECISION 32
#define HIGH_RES_PRECISION 64
#define ZOOM_INCREMENT 2
#define FINAL_PRECISION HIGH_RES_PRECISION+ZOOM_INCREMENT*ZOOM_LEVEL

/* png related values */

#define WIDTH  320//1024
#define HEIGHT 200//768

#define LOW_RES_WIDTH 20
#define LOW_RES_HEIGHT 20

#define RANDOM 0

#define GARBAGE_COLLECTION 0
#define GARBAGE_SIZE 20000

#define INITIAL_CRC 0xFFFFFFFFL

/* note: this will evaluate x twice, which doesn't matter in this case */
#define absolute(x) ( (x) < 0 ? -(x) : (x) )


extern void build_fractal(FRACTAL *f);
extern int pick_quarter(FRACTAL *f);

// from imlib-stuff.c
extern void write_image_to_file(FRACTAL *, const char *);

// from period.cc
extern unsigned char *generate_MD5_key(COMPLEX_MPF *, unsigned char *);
extern void free_digest_tree(DIGEST_TREE_NODE *);
extern DIGEST_TREE_NODE *new_digest_node(unsigned char *);
extern unsigned char add_to_periodicity_tree(COMPLEX_MPF *);
extern long calc_height(DIGEST_TREE_NODE *t);
extern void init_garbage();
extern void init_lookup();
extern void free_all_garbage();
extern void free_all_trees();
//extern gboolean lookup_and_search(unsigned char *d);

// from crc-32.c
extern void gen_crc_table();
extern unsigned long update_crc(unsigned long crc_accum, char *data_blk_ptr, int data_blk_size);

// file-io.c

extern int write_fractal_to_file(FRACTAL *f, char *fname);
extern int read_fractal_from_file(FRACTAL *f, char *fname);









