#include <cstring>

#ifndef NDEBUG
#include <iostream>
#include <cassert>
#endif

#include "trie.h"

static trieNode *shithole = 0;

loSatTN::loSatTN(char d, trieNode **p) : trieNode(d) {
	   ptr = p;
	   v[0] = v[1] = v[2] = -1;
}

loSatTN::~loSatTN() {
		if (v[0] >= 0) delete c[0];
		if (v[1] >= 0) delete c[1];
		if (v[2] >= 0) delete c[2];
}
	
trieNode *loSatTN::add(char *str, int len, int val) {
		int l = str[0];
#ifndef NDEBUG
		assert(l >= 0);
#endif
		if (len == 0) {
		     code = val;
		     return this;
		}
		
		if (v[0] < 0) {
		     v[0] = l;
		     c[0] = new loSatTN(depth + 1, &c[0]);
		     return c[0]->add(&str[1], len - 1, val);
		}

		if (v[0] == l) {
		     return c[0]->add(&str[1], len - 1, val);
		} 
		
		if (v[1] < 0) {
		     v[1] = l;
		     c[1] = new loSatTN(depth + 1, &c[1]);
		     return c[1]->add(&str[1], len - 1, val);
		}
		
		if (v[1] == l) {
		     return c[1]->add(&str[1], len - 1, val);
		} 
		
		if (v[2] < 0) {
		     v[2] = l;
		     c[2] = new loSatTN(depth + 1, &c[2]);
		     return c[2]->add(&str[1], len - 1, val);
		} 
		
		if (v[2] == l) {
		     return c[2]->add(&str[1], len - 1, val);
		}

		     static trieNode *n = new hiSatTN(*this);
#ifndef NDEBUG
		     cerr << "Lo to Hi conversion." << endl;
		     cerr << val << " as [" << len << "]";
		     for (int i = 0; i < len; i++) {
			cerr << str[i];
		     }
		     cerr << endl;
#endif
		     n->add(str, len, val);
		     ptr[0] = n;
		     shithole = this;
		     return n;
}
	
trieNode *loSatTN::lookup(char *str, int len) {
		char l = str[0];
		if (len == 0) return this;
		if (v[0] == l) return c[0]->lookup(&str[1], len - 1);
		if (v[1] == l) return c[1]->lookup(&str[1], len - 1);
		if (v[2] == l) return c[2]->lookup(&str[1], len - 1);
		return this;
}

hiSatTN::hiSatTN(char d) : trieNode(d) {
   memset(c, 0, sizeof(trieNode *[256]));
}

hiSatTN::hiSatTN(loSatTN &tn) : trieNode(tn.depth) {
   memset(c, 0, sizeof(trieNode *[256]));
   if (tn.v[0] >= 0) {
      c[tn.v[0]] = tn.c[0];
      tn.v[0] = -1;
      if (tn.v[1] >= 0) {
	 c[tn.v[1]] = tn.c[1];
	 tn.v[1] = -1;
	 if (tn.v[2] >= 0) {
	    c[tn.v[2]] = tn.c[2];
	    tn.v[2] = -1;
	 }
      }
   }
}

trieNode * hiSatTN::add(char *str, int len, int val) {
   int l = str[0];
#ifndef NDEBUG
   assert(l >= 0);

    cerr << "adding " << val << " as [" << len << "]";
    for (int i = 0; i < len; i++) {
    	cerr << str[i];
    }
    cerr << endl;
#endif

   if (len == 0) {
      code = val;
      return this;
   } else if (c[l]) {
      return c[l]->add(&str[1], len - 1, val);
   } else {
      c[l] = new loSatTN(depth + 1, &c[l]);
      return c[l]->add(&str[1], len - 1, val);
   }
}
	
trieNode *hiSatTN::lookup(char *str, int len) {
   char l = str[0];

   if (len == 0) return this;

#ifndef NDEBUG
   assert(l >= 0);
#endif
   if (c[l]) return c[l]->lookup(&str[1], len - 1);
   return this;
}

/*
	   if ((llen < 21) && (token_count < token_max)) {
	      last[llen] = cur[0];
	      trie.add(last, llen, token_count++);
	      if (shithole) {
		 delete shithole;
		 shithole = NULL;
	      }
	   }
*/
