#include "deep_zoom.h"

// recursively free the nodes of this tree

void free_color_tree(COLOR_TREE_NODE *t)
{
  if(t==NULL)
    return;

  free_color_tree(t->left);
  free_color_tree(t->right);
  free(t);
}

// calculate which quadrant of the image (x, y) is in

int find_which_quadrant(FRACTAL *f, int x, int y)
{
  if(x<f->width/2)
    {
      if(y<f->height/2)
	{
	  return(1);
	}
      else
	{
	  return(2);
	}
    }
  else
    {
      if(y<f->height/2)
	{
	  return(0);
	}
      else
	{
	  return(3);
	}
    }
  return(1);
}

// creates a new node for the color tree

COLOR_TREE_NODE *new_color_node(DATA32 d)
{
  COLOR_TREE_NODE *ret;

  ret=(COLOR_TREE_NODE *) malloc(sizeof(COLOR_TREE_NODE));
  ret->left=NULL;
  ret->right=NULL;
  ret->data=d;
  ret->color_counter=1;
  return(ret);
}

unsigned char find_or_add_to_color_tree(COLOR_TREE_NODE *t, DATA32 d)
{
  // special case for the root node

  if((signed) t->data==-1)
    {
      t->data=d;
      return(FALSE);
    }

  if(t->data==d)
    {
      ++t->color_counter;
      return(TRUE);
    }

  if(t->data>d)
    {
      if(t->left!=NULL)
	{
	  return(find_or_add_to_color_tree(t->left, d));
	}
      t->left=new_color_node(d);
      return(FALSE);
    }


  if(t->right!=NULL)
    {
      return(find_or_add_to_color_tree(t->right, d));
    }
  t->right=new_color_node(d);

  return(FALSE);
}

#if 0
unsigned long x_y_to_offset(FRACTAL *f, unsigned long x, unsigned long y)
{
  return(f->width*y+x);
}
#else
#define x_y_to_offset(f,x,y) ((f)->width*(y)+(x))
#endif

// recursively calculate the number of unique colors in an image (counts the number of nodes in the
// binary tree)

unsigned long calc_active_colors(COLOR_TREE_NODE *n)
{
  unsigned long counter=0;

  if(n->left!=NULL)
    counter+=calc_active_colors(n->left);
  if(n->right!=NULL)
    counter+=calc_active_colors(n->right);
  return(counter+1);
}

// this function selects which quarter to zoom into next

int pick_quarter(FRACTAL *f)
{
  int i, j, q;
  COLOR_TREE_NODE *c_trees[4];

  unsigned long high, high_index, active_colors[4];
  unsigned long random_index;

  // create four new binary trees to store the colors
  
  // the unsigned thing is a pretty bad hack

  c_trees[0]=new_color_node((unsigned)-1);
  c_trees[1]=new_color_node((unsigned)-1);
  c_trees[2]=new_color_node((unsigned)-1);
  c_trees[3]=new_color_node((unsigned)-1);

  for(i=0;i<4;++i)
      active_colors[i]=0;

  // add each pixel to the binary search tree

  for(i=0;i<f->width;++i)
    {
      for(j=0;j<f->height;++j)
	{
	  q=find_which_quadrant(f, i, j);
	  find_or_add_to_color_tree(c_trees[q], 
				    f->image_data[x_y_to_offset(f, i, j)]);
	}
    }

  // calculate how many unique colors each quadrant has and free the tree used to calculate it

  for(i=0;i<4;++i)
    {
      active_colors[i]=calc_active_colors(c_trees[i]);
      printf("%ld ", active_colors[i]);
      free_color_tree(c_trees[i]);
    }

  printf("\n");
  high=active_colors[0];
  high_index=0;

  // find which quadrant has the most unique colours in it

  for(i=0;i<4;++i)
    {
      if(active_colors[i]>high)
	{
	  high_index=i;
	  high=active_colors[i];
	}
    }

/*      to avoid having the program zoom into an idential area each time, the  */
/*      program picks a random quadrant and if that quadrant has a score 80%  */
/*      as high as the highest score that quadrant is selected instead */

  if(f->options.random)
    {
      random_index=rand()%4;
      if(high*0.8<active_colors[random_index])
	{
	  high_index=random_index;
	  high=active_colors[random_index];
	}
    }

  printf("high_index=%ld high = %ld\n", high_index, high);

  // return the quadrant to zoom into
  return(high_index);
}






